"""Defines and handles the Resource object returned by the API."""
import requests

from fabman.fabman_object import FabmanObject


class ResourceBridge(FabmanObject):
    """Simple class to hold Bridge information for a Bridge connected to a resource.
    Further used to issue commands to the bridge
    """

    def __str__(self):
        return f"Resource #{self.resource_id}: {self.serialNumber}"

    def update(self, **kwargs) -> None:
        """
        Update the bridge object on the server. Returns the updated bridge object.
        Updates all information in place.

        :calls: "PUT /resources/{resource_id}/bridge" \
		<https://fabman.io/api/v1/documentation#/resources/putResourcesIdBridge>
  
        :return: None -- attributes are updated in place
        :rtype: None
        """
        uri = f"/resources/{self.resource_id}/bridge"

        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)


class ResourceBridgeApiKey(FabmanObject):
    """Simple class to hold Bridge API key information"""

    def __str__(self):
        api_key = "xxxx" + self.apiKey[-4:]
        return f"Resource #{self.resource_id}: {api_key}"


class Resource(FabmanObject):
    """
    Resource object returned by the API. Provides access to all API calls that
    operate on a single Resource.
    """

    def __str__(self):
        return f"Resource #{self.id}: {self.name}"

    def delete_bridge(self, **kwargs) -> requests.Response:
        """
        Delete the bridge that is currently connected to this resource. If no bridge
        is currently connected, this will do nothing.

        :calls: "DELETE /resources/{id}/bridge" \
		<https://fabman.io/api/v1/documentation#/resources/deleteResourcesIdBridge>
  
        :return: response information of call
        :rtype: requests.Response
        """
        uri = f"/resources/{self.id}/bridge"

        response = self._requester.request("DELETE", uri, _kwargs=kwargs)

        return response

    def delete(self, **kwargs) -> requests.Response:
        """
        Deletes the resource. *WARNING: THIS CANNOT BE UNDONE.* All future API
        calls from this resource will fail.

        :calls: "DELETE /resources/{id}" \
		<https://fabman.io/api/v1/documentation#/resources/deleteResourcesId>
  
        :return: response information of call
        :rtype: requests.Response
        """
        uri = f"/resources/{self.id}"

        response = self._requester.request("DELETE", uri, _kwargs=kwargs)

        return response

    def get_bridge(self, **kwargs) -> ResourceBridge:
        """
        Get the bridge that is currently connected to this resource. If no bridge is
        connected, this will return an empty list.

        :calls: "GET /resources/{id}/bridge" \
		<https://fabman.io/api/v1/documentation#/resources/getResourcesIdBridge>
  
        :return: ResourceBridge object with information on the bridge
        :rtype: fabman.resource.ResourceBridge
        
        """
        uri = f"/resources/{self.id}/bridge"

        response = self._requester.request("GET", uri, _kwargs=kwargs)

        data = response.json()
        data.update({"resource_id": self.id})

        return ResourceBridge(self._requester, data)

    def get_bridge_api_key(self, **kwargs) -> ResourceBridgeApiKey:
        """
        Get the API key for the bridge that is currently connected to this resource.
        For most users, this will return a 204 code with an empty body. Only superusers
        or users who have created a custom bridge should be able to access this endpoint.

        :calls: "GET /resources/{id}/bridge/api-key" \
		<https://fabman.io/api/v1/documentation#/resources/getResourcesIdBridgeApikey>

        :return: response information of call
        :rtype: ResourceBridgeApiKey
        """
        uri = f"/resources/{self.id}/bridge/api-key"

        response = self._requester.request("GET", uri, _kwargs=kwargs)

        data = response.json()
        data.update({"resource_id": self.id})

        return ResourceBridgeApiKey(self._requester, data)

    def switch_on(self, **kwargs) -> requests.Response:
        """
        Switch on the resource. Requires "code" field to be in the data packet.
        Does not seem to work without the proper code, which is undocumented.

        :calls: "POST /resources/{id}/switch-on" \
		<https://fabman.io/api/v1/documentation#/resources/postResourcesIdBridgeSwitchon>
  
        :return: response information of call
        :rtype: requests.Response
        """
        uri = f"/resources/{self.id}/switch-on"

        response = self._requester.request("POST", uri, _kwargs=kwargs)

        return response

    def update(self, **kwargs) -> None:
        """
        Update the resource.

        :calls: "PUT /resources/{id}" \
		<https://fabman.io/api/v1/documentation#/resources/putResourcesId>
  
        :return: None -- attributes are updated in place
        :rtype: None
        """
        uri = f"/resources/{self.id}"

        kwargs.update({"lockVersion": self.lockVersion})

        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)



"""Declares the Job object"""

from fabman.fabman_object import FabmanObject


class Job(FabmanObject):
    """
    Simple object for handling Jobs. No methods are currently available for Jobs
    """

    def __str__(self):
        return f"Job #{self.id}"



"""Defines the Package class"""

import requests

from fabman.fabman_object import FabmanObject
from fabman.paginated_list import PaginatedList


class PackageCredit(FabmanObject):
    """Simple Class to handle PackageCredits"""

    def __str__(self):
        return f"PackageCredit #{self.id}, Package #{self.package_id}: {self.scope}"

    def delete(self, **kwargs) -> requests.Response:
        """
        Deletes a credit from a package. *WARNING: This is irreversible.*

        :calls: "DELETE /packages/{packageId}/credits/{creditId}" \
		<https://fabman.io/api/v1/documentation#/packages/deletePackagesIdCreditsCreditid>
  
        :returns: response information of call
        :rtype: requests.Response
        """

        uri = f"/packages/{self.package_id}/credits/{self.id}"

        response = self._requester.request("DELETE", uri, _kwargs=kwargs)

        return response

    def update(self, **kwargs) -> None:
        """
        Updates the credit. Attributes are updated in place with new information.

        :calls: "PUT /packages/{packageId}/credits/{creditId}" \
		<https://fabman.io/api/v1/documentation#/packages/putPackagesIdCreditsCreditid>

        :return: None -- attributes are updated in place
        :rtype: None
        """

        uri = f"/packages/{self.package_id}/credits/{self.id}"

        kwargs.update({"lockVersion": self.lockVersion})
        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)


class PackagePermission(FabmanObject):
    """Simple Class to handle PackagePermissions"""

    def __str__(self):
        return f"PackagePermission #{self.id}, Package #{self.package_id}: {self.type}"

    def delete(self, **kwargs) -> requests.Response:
        """
        Deletes a permission from a package. *WARNING: This is irreversible.*

        :calls: "DELETE /packages/{packageId}/permissions/{permissionId}" \
		<https://fabman.io/api/v1/documentation#/packages/deletePackagesIdPermissionsPermissionid>

        :return: response information of call
        :rtype: requests.Response
        """

        uri = f"/packages/{self.package_id}/permissions/{self.id}"

        response = self._requester.request("DELETE", uri, _kwargs=kwargs)

        return response

    def update(self, **kwargs) -> None:
        """
        Updates the permission. Attributes are updated in place with new information.

        :calls: "PUT /packages/{packageId}/permissions/{permissionId}" \
		<https://fabman.io/api/v1/documentation#/packages/putPackagesIdPermissionsPermissionid>
  
        :return: None -- attributes are updated in place
        :rtype: None
        """

        uri = f"/packages/{self.package_id}/permissions/{self.id}"

        kwargs.update({"lockVersion": self.lockVersion})
        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)


class Package(FabmanObject):
    """Handles all interaction with Package objects on the Fabman API"""

    def __str__(self):
        return f"Package #{self.id}: {self.name}"

    def create_credit(self, **kwargs) -> PackageCredit:
        """
        Create a credit for this package. Takes no arguments.

        :calls: "POST /packages/{id}/credits" \
		<https://fabman.io/api/v1/documentation#/packages/postPackagesIdCredits>

        :return: Object with information and methods of the new credit
        :rtype: fabman.package.PackageCredit
        """

        uri = f"/packages/{self.id}/credits"
        response = self._requester.request("POST", uri, _kwargs=kwargs)

        data = response.json()
        data.update({"package_id": self.id})

        return PackageCredit(self._requester, data)

    def create_permission(self, **kwargs) -> PackagePermission:
        """
        Create a new permission for this package. Gives new abilities to package hodlers.

        :calls: "POST /packages/{id}/permissions" \
		<https://fabman.io/api/v1/documentation#/packages/postPackagesIdPermissions>
  
        :return: Object with information and methods of the new permission
        :rtype: fabman.package.PackagePermission
        """

        uri = f"/packages/{self.id}/permissions"

        response = self._requester.request("POST", uri, _kwargs=kwargs)

        data = response.json()
        data.update({"package_id": self.id})

        return PackagePermission(self._requester, data)

    def get_credit(self, credit_id, **kwargs) -> PackageCredit:
        """
        Return a credit for this package given a credit_id.

        :calls: "GET /packages/{id}/credits/{creditId}" \
		<https://fabman.io/api/v1/documentation#/packages/getPackagesIdCreditsCreditid>

        :param credit_id: ID of the credit to retrieve
        :type credit_id: int
        :return: Object with information and methods of the credit
        :rtype: fabman.package.PackageCredit
        """

        uri = f"/packages/{self.id}/credits/{credit_id}"
        response = self._requester.request("GET", uri, _kwargs=kwargs)

        data = response.json()
        data.update({"package_id": self.id})

        return PackageCredit(self._requester, data)

    def get_credits(self, **kwargs) -> PaginatedList:
        """
        Return a PaginatedList of credits for this package

        Calls: "GET /packages/{id}/credits"
        Documentation: https://fabman.io/api/v1/documentation#/packages/getPackagesIdCredits

        :return: List of credits for this package
        :rtype: fabman.paginated_list.PaginatedList
        """

        return PaginatedList(
            PackageCredit,
            self._requester,
            "GET",
            f"/packages/{self.id}/credits",
            extra_attribs={"package_id": self.id},
            kwargs=kwargs,
        )

    def get_permission(self, permission_id, **kwargs) -> PackagePermission:
        """
        Return a permission for this package given a permission_id.

        :calls: "GET /packages/{id}/permissions/{permissionId}" \
		<https://fabman.io/api/v1/documentation#/packages/getPackagesIdPermissionsPermissionid>
  
        :returns: Object with information and methods of the permission
        :rtype: fabman.package.PackagePermission
        """
        uri = f"/packages/{self.id}/permissions/{permission_id}"
        response = self._requester.request("GET", uri, _kwargs=kwargs)

        data = response.json()
        data.update({"package_id": self.id})

        return PackagePermission(self._requester, data)

    def get_permissions(self, **kwargs) -> PaginatedList:
        """
        Return a PaginatedList of permissions for the package.

        :calls: "GET /packages/{id}/permissions" \
		<https://fabman.io/api/v1/documentation#/packages/getPackagesIdPermissions>

        :returns: List of permissions for the package
        :rtype: fabman.paginated_list.PaginatedList
        """

        return PaginatedList(
            PackagePermission,
            self._requester,
            "GET",
            f"/packages/{self.id}/permissions",
            extra_attribs={"packageId": self.id},
            kwargs=kwargs,
        )

    def delete(self, **kwargs) -> requests.Response:
        """
        Deletes the package. Takes no arguments. **WARNINGL: This is irreversible.**

        :calls: "DELETE /packages/{id}" \
		<https://fabman.io/api/v1/documentation#/packages/deletePackagesId>
  
        :returns: response information of call
        :rtype: requests.Response
        """

        uri = f"/packages/{self.id}"

        response = self._requester.request("DELETE", uri, _kwargs=kwargs)

        return response

    def update(self, **kwargs) -> None:
        """
        Update information on the package. Note that many fields may be unalterable
        by the API key holder.

        :calls: "PUT /packages/{id}" \
		<https://fabman.io/api/v1/documentation#/packages/putPackagesId>
  
        :return: None -- attributes are updated in place
        :rtype: None
        """

        uri = f"/packages/{self.id}"

        kwargs.update({"lockVersion": self.lockVersion})

        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)



"""Sanity Checks"""
import unittest

import fabman


class SanityTest(unittest.TestCase):
    """Quick sanity check to make sure tests are working fine"""

    def test_sanity(self):
        """Quick sanity check to make sure tests are working fine"""
        self.assertEqual(1, 1)

    def test_fabman_import(self):
        """Quick sanity to check to ensure fabman is imported correctly"""
        key = "123"
        fab = fabman.Fabman(key)
        self.assertEqual(fab.__class__.__name__, "Fabman")


if __name__ == "__main__":
    unittest.main()



"""Defines the ResourceType class"""

import requests

from fabman.fabman_object import FabmanObject


class ResourceType(FabmanObject):
    """Class for interacting with the resource-types endpoint in the Fabman API"""

    def __str__(self):
        return f"ResourceType #{self.id}: {self.name}"

    def delete(self, **kwargs) -> requests.Response:
        """
        Deletes a resource-type. *WARNING: This is irreversible and may have
        unintended consequences.* Resource Types are used to group resources and
        are therefore tied to all resources of that type. Be sure you know
        what you are doing when using this endpoint. There is no confirmation.

        :calls: "DELETE /resource-types/{resourceTypeId}" \
		<https://fabman.io/api/v1/documentation#/resource-types/deleteResourcetypesId>
  
        :return: response information of call
        :rtype: requests.Response
        """

        uri = f"/resource-types/{self.id}"

        response = self._requester.request("DELETE", uri, _kwargs=kwargs)

        return response

    def update(self, **kwargs) -> None:
        """
        Updates the resource-type. Attributes are updated in place with new information
        retrieved from the API.

        :calls: "PUT /resource-types/{resourceTypeId}" \
		<https://fabman.io/api/v1/documentation#/resource-types/putResourcetypesId>

        :return: None -- attributes are updated in place
        :rtype: None
        """

        uri = f"/resource-types/{self.id}"

        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)



"""Exceptions for use throughout the package"""


class FabmanException(Exception):
    """Base class for all Fabman exceptions"""

    def __init__(self, message):
        if isinstance(message, dict):
            errors = message.get("errors", None)

            if errors:
                self.message = errors
            else:
                self.message("Something went wrong ", message)
        else:
            self.message = message

    def __str__(self):
        return str(self.message)


class BadRequest(FabmanException):
    """
    Fabman was unable to understand the request. More information may be needed or
    arguments may need to be fixed.
    """


class InvalidAccessToken(FabmanException):
    """
    The access token provided was invalid.
    """


class Unauthorized(FabmanException):
    """
    The access token provided was not authorized to access the resource.
    """


class ResourceDoesNotExist(FabmanException):
    """
    The resource requested does not exist.
    """


class ForbiddenError(FabmanException):
    """
    The access token provided does not have permission to access the resource.
    """


class Conflict(FabmanException):
    """
    The request could not be completed due to a conflict with the current
    state of the resource.
    """


class UnprocessableEntity(FabmanException):
    """
    The request was well-formed but was unable to be followed due to semantic errors.
    """


class RateLimitExceeded(FabmanException):
    """
    The request was valid, but too may requests have been issued from this access token.
    Please try again later.
    """



"""Tests for the Resources class."""
# pylint: disable=missing-function-docstring, missing-class-docstring, invalid-name, unused-argument

import unittest

import requests
import requests_mock

from fabman import Fabman
from fabman.resource import Resource, ResourceBridge, ResourceBridgeApiKey
from tests import settings
from tests.util import register_uris, validate_update


@requests_mock.Mocker()
class TestResource(unittest.TestCase):
    """Test Cases for the Resource class."""

    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"fabman": ["get_resource_by_id"]}, m)

            self.resource: Resource = self.fabman.get_resource(1)

    def test_instance(self, m):
        self.assertIsInstance(self.resource, Resource)

    def test_str(self, m):
        string = str(self.resource)
        self.assertTrue(string == "Resource #1: USS Defiant")

    def test_delete(self, m):
        register_uris({"resource": ["delete"]}, m)

        resp = self.resource.delete()
        self.assertTrue(m.called)
        self.assertIsInstance(resp, requests.Response)
        self.assertTrue(resp.status_code == 204)

    def test_delete_bridge(self, m):
        register_uris({"resource": ["delete_bridge"]}, m)

        resp = self.resource.delete_bridge()
        self.assertTrue(m.called)
        self.assertIsInstance(resp, requests.Response)
        self.assertTrue(resp.status_code == 204)

    def test_get_bridge(self, m):
        register_uris({"resource": ["get_bridge"]}, m)
        bridge = self.resource.get_bridge()

        self.assertIsInstance(bridge, ResourceBridge)
        self.assertTrue(hasattr(bridge, "serialNumber"))

    def test_get_bridge_api_key(self, m):
        register_uris({"resource": ["get_bridge_api_key"]}, m)
        api_key = self.resource.get_bridge_api_key()

        self.assertIsInstance(api_key, ResourceBridgeApiKey)
        self.assertTrue(api_key.resource_id == 1)

        self.assertEqual(str(api_key), "Resource #1: xxxx3456")

    def test_switch_on(self, m):
        register_uris({"resource": ["switch_on"]}, m)

        resp = self.resource.switch_on()
        self.assertTrue(m.called)
        self.assertIsInstance(resp, requests.Response)
        self.assertTrue(resp.status_code == 201)

    def test_update(self, m):
        m.register_uri(
            "PUT",
            f"{settings.BASE_URL_WITH_VERSION}/resources/1",
            text=validate_update,
            status_code=200,
        )

        self.resource.update(title="USS Voyager")
        self.assertTrue(m.called)


@requests_mock.Mocker()
class TestResourceBridge(unittest.TestCase):
    """Tests the ResourceBridge class"""

    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris(
                {"fabman": ["get_resource_by_id"], "resource": ["get_bridge"]}, m
            )

            self.resource = self.fabman.get_resource(1)
            self.bridge = self.resource.get_bridge()

    def test_instance(self, m):
        self.assertIsInstance(self.bridge, ResourceBridge)

    def test_str(self, m):
        string = str(self.bridge)
        self.assertTrue(string == "Resource #1: abcdef123456")

    def test_update(self, m):
        """This update does not require a lockVersion for whatever reason"""

        register_uris({"resource": ["update_bridge"]}, m)

        self.bridge.update()
        self.assertTrue(m.called)



"""Utility Function tests"""
# pylint: disable=missing-docstring, invalid-name, unused-argument
import unittest

import requests_mock

from fabman.util import clean_headers

# pylint: disable=missing-class-docstring, missing-function-docstring, too-many-public-methods


@requests_mock.Mocker()
class TestUtil(unittest.TestCase):
    def test_clean_headers_no_auth(self, m):
        headers = {
            "a": "b",
        }
        out = clean_headers(headers)
        self.assertDictEqual(headers, out)

    def test_clean_headers_auth(self, m):
        headers = {"Authorization": "thisisatoken"}
        out = clean_headers(headers)
        self.assertEqual(out["Authorization"], "****oken")



"""Tests for the PaginatedList class."""
# pylint: disable=missing-docstring, invalid-name, unused-argument

import unittest

import requests_mock

from fabman import Fabman
from fabman.member import Member
from fabman.paginated_list import PaginatedList
from tests import settings
from tests.util import register_uris


@requests_mock.Mocker()
class TestPaginatedList(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"fabman": ["get_members"]}, m)

            self.paginated_list: PaginatedList = self.fabman.get_members(limit=1)

    def test_sanity(self, m):
        self.assertEqual(1, 1)

    def test_repr(self, m):
        self.assertEqual("<PaginatedList of type Member>", repr(self.paginated_list))

    def test_format_link(self, m):
        headers = {
            "link": '<https://fabman.io/api/v1/members?limit=1&offset=1>; rel="next", <https://fabman.io/api/v1/members?limit=1&offset=2>; rel="last"'
        }
        link = self.paginated_list._PaginatedList__format_link(headers)
        self.assertEqual(link, "/members?limit=1&offset=1")

    def test_iteration(self, m):
        register_uris(
            {
                "paginated_list": ["get_members_first", "get_members_second"],
            },
            m,
        )

        members = self.fabman.get_members(limit=5)
        self.assertIsInstance(members, PaginatedList)
        items = [m for m in members]
        self.assertEqual(len(items), 10)
        for i, item in enumerate(items):
            self.assertIsInstance(item, Member)
            self.assertEqual(item.id, i + 1)

    def test_negative_index(self, m):
        register_uris({"paginated_list": ["get_members_first"]}, m)

        members = self.fabman.get_members(limit=5)
        self.assertIsInstance(members, PaginatedList)

        with self.assertRaises(IndexError):
            member = members[-1]

    def test_index(self, m):
        register_uris({"paginated_list": ["get_members_first"]}, m)

        members = self.fabman.get_members(limit=5)
        self.assertIsInstance(members, PaginatedList)
        member = members[2]
        self.assertIsInstance(member, Member)
        self.assertTrue(member.id == 3)

    def test_grow_index(self, m):
        register_uris(
            {"paginated_list": ["get_members_first", "get_members_second"]}, m
        )

        members = self.fabman.get_members(limit=5)
        self.assertIsInstance(members, PaginatedList)
        member = members[7]
        self.assertIsInstance(member, Member)
        self.assertTrue(member.id == 8)

    def test_index_out_of_range(self, m):
        register_uris(
            {"paginated_list": ["get_members_first", "get_members_second"]}, m
        )

        members = self.fabman.get_members(limit=5)
        self.assertIsInstance(members, PaginatedList)

        with self.assertRaises(IndexError):
            member = members[11]



"""Tests for the Webhook Class"""
# pylint: disable=missing-docstring, invalid-name, unused-argument

import unittest

import requests
import requests_mock

from fabman import Fabman
from fabman.webhook import Webhook
from tests import settings
from tests.util import register_uris, validate_update


@requests_mock.Mocker()
class TestWebhook(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"fabman": ["get_webhook_by_id"]}, m)

            self.webhook: Webhook = self.fabman.get_webhook(1)

    def test_instance(self, m):
        self.assertIsInstance(self.webhook, Webhook)

    def test_str(self, m):
        string = str(self.webhook)
        self.assertTrue(string == "Webhook #1: Garak's Test Webhook")

    def test_delete(self, m):
        register_uris({"webhook": ["delete"]}, m)

        response = self.webhook.delete()
        self.assertTrue(m.called)
        self.assertIsInstance(response, requests.Response)
        self.assertTrue(response.status_code == 204)

    def test_get_events(self, m):
        register_uris({"webhook": ["get_events"]}, m)

        resp = self.webhook.get_events()
        self.assertTrue(m.called)
        self.assertIsInstance(resp, requests.Response)
        self.assertTrue(resp.status_code == 200)

    def test_send_test_event(self, m):
        register_uris({"webhook": ["send_test_event"]}, m)

        resp = self.webhook.send_test_event()
        self.assertTrue(m.called)
        self.assertIsInstance(resp, requests.Response)
        self.assertTrue(resp.status_code == 201)

    def test_update(self, m):
        m.register_uri(
            "PUT",
            f"{settings.BASE_URL_WITH_VERSION}/webhooks/1",
            text=validate_update,
            status_code=200,
        )

        self.webhook.update(label="New Label")
        self.assertTrue(m.called)



"""Defines the ResourceLog class"""

import requests

from fabman.fabman_object import FabmanObject


class ResourceLog(FabmanObject):
    """Class for interacting with the resource-logs endpoint on the Fabman API"""

    def __str__(self):
        return f"ResourceLog #{self.id}, Resource #{self.resource} - {self.type}"

    def delete(self, **kwargs) -> requests.Response:
        """
        Deletes a resource-log. *WARNING: This is irreversible.*

        :calls: "DELETE /resource-logs/{resourceLogId}" \
		<https://fabman.io/api/v1/documentation#/resource-logs/deleteResourcelogsId>
  
        :return: response information of call
        :rtype: requests.Response
        """

        uri = f"/resource-logs/{self.id}"

        response = self._requester.request("DELETE", uri, _kwargs=kwargs)

        return response

    def update(self, **kwargs) -> None:
        """
        Updates the resource-log. Attributes are updated in place with new information
        retrieved from the API.

        :calls: "PUT /resource-logs/{resourceLogId}" \
		<https://fabman.io/api/v1/documentation#/resource-logs/putResourcelogsId>
  
        :return: None -- attributes are updated in place
        :rtype: None
        """

        uri = f"/resource-logs/{self.id}"

        kwargs.update({"lockVersion": self.lockVersion})

        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)



"""Tests for the ResourceLog class."""
# pylint: disable=missing-docstring, invalid-name, unused-argument

import unittest

import requests
import requests_mock

from fabman import Fabman
from fabman.resource_log import ResourceLog
from tests import settings
from tests.util import register_uris, validate_update


@requests_mock.Mocker()
class TestResourceLog(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"fabman": ["get_resource_log_by_id"]}, m)

            self.resource_log: ResourceLog = self.fabman.get_resource_log(1)

    def test_instance(self, m):
        self.assertIsInstance(self.resource_log, ResourceLog)

    def test_str(self, m):
        string = str(self.resource_log)
        self.assertEqual(string, "ResourceLog #1, Resource #1 - reboot")

    def test_delete(self, m):
        register_uris({"resource_log": ["delete"]}, m)

        resp = self.resource_log.delete()
        self.assertTrue(m.called)
        self.assertIsInstance(resp, requests.Response)
        self.assertTrue(resp.status_code == 204)

    def test_update(self, m):
        m.register_uri(
            "PUT",
            f"{settings.BASE_URL_WITH_VERSION}/resource-logs/1",
            text=validate_update,
            status_code=200,
        )

        self.resource_log.update(type="reboot")
        self.assertTrue(m.called)



"""Tests for the Invoice class."""
# pylint: disable=missing-docstring, invalid-name, unused-argument

import unittest

import requests_mock

from fabman import Fabman
from fabman.invoice import Invoice, InvoiceDetails
from tests import settings
from tests.util import register_uris, validate_update


@requests_mock.Mocker()
class TestInvoice(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"fabman": ["get_invoice_by_id"]}, m)

            self.invoice: Invoice = self.fabman.get_invoice(1)

    def test_instance(self, m):
        self.assertIsInstance(self.invoice, Invoice)

    def test_str(self, m):
        string = str(self.invoice)

        self.assertTrue(string == "Invoice #1: 12.34 unpaid")

    def test_cancel(self, m):
        m.register_uri(
            "POST",
            f"{settings.BASE_URL_WITH_VERSION}/invoices/1/cancel",
            text=validate_update,
            status_code=200,
        )

        self.invoice.cancel()
        self.assertTrue(m.called)

    def test_details(self, m):
        register_uris({"invoice": ["details"]}, m)

        details = self.invoice.details()
        self.assertTrue(m.called)
        self.assertIsInstance(details, InvoiceDetails)
        self.assertTrue(details.id == 1)

        self.assertEqual(str(details), "Invoice #1: 12.34 # Charges: 1")

    def test_details_embedded(self, m):
        register_uris({"fabman": ["get_invoice_embedded_detail"]}, m)
        invoice = self.fabman.get_invoice(1)

        details = invoice.details()
        self.assertTrue(m.called)
        self.assertIsInstance(details, InvoiceDetails)
        self.assertEqual(details.id, 2)

        self.assertEqual(str(details), "Invoice #2: 45.67 # Charges: 1")

    def test_update(self, m):
        m.register_uri(
            "PUT",
            f"{settings.BASE_URL_WITH_VERSION}/invoices/1",
            text=validate_update,
            status_code=200,
        )

        self.invoice.update()
        self.assertTrue(m.called)



"""Handles the Charge Object for managing fabman charges"""

import requests

from fabman.fabman_object import FabmanObject


class Charge(FabmanObject):
    """Charge object handles management of charges in fabman"""

    def __str__(self):
        return f"Charge #{self.id}: {self.price} {self.description}"

    def delete(self, **kwargs) -> requests.Response:
        """
        Deletes the charge. *WARNING: THIS CANNOT BE UNDONE.*

        :calls: "DELETE /charges/{id}" \
		<https://fabman.io/api/v1/documentation#/charges/deleteChargesId>
  
        :returns: Response from the API with status code 204 (No Content) if successful
        :rlist: requests.Response
        """

        response = self._requester.request(
            "DELETE", f"/charges/{self.id}", _kwargs=kwargs
        )

        return response

    def update(self, **kwargs) -> None:
        """
        Update the charge object on the server. Updates all information in place
        with returned data from the server.

        :calls: "PUT /charges/{id}" \
		<https://fabman.io/api/v1/documentation#/charges/putChargesId>
  
        :returns: None
        :rlist: None
        """

        kwargs.update({"lockVersion": self.lockVersion})

        response = self._requester.request("PUT", f"/charges/{self.id}", _kwargs=kwargs)

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)



"""Tests for the Jobs class."""
# pylint: disable=missing-docstring, invalid-name, unused-argument

import unittest

import requests_mock

from fabman import Fabman
from fabman.job import Job
from tests import settings
from tests.util import register_uris


@requests_mock.Mocker()
class TestMembers(unittest.TestCase):
    def setUp(self):
        self.fabman = Fabman(settings.API_KEY)

        with requests_mock.Mocker() as m:
            register_uris({"fabman": ["get_job_by_id"]}, m)

            self.job: Job = self.fabman.get_job(1)

    def test_sanity(self, m):
        self.assertEqual(1, 1)



"""Defines the Account object"""

from fabman.fabman_object import FabmanObject


class PaymentInfo(FabmanObject):
    """Holds PaymentInfo as returned from the accounts endpoint."""

    def __str__(self):
        return f"PaymentInfo for Account #{self.account_id}: {self.paymentMethod}"


class Account(FabmanObject):
    """Handles Account objects from the Fabman API. In most cases, this only
    returns information about the API key holders account. Since most API endpoints
    in this object require superuser privileges, they are not implemented.
    """

    def __str__(self):
        return f"Account #{self.id}: {self.name}"

    def update(self, **kwargs) -> None:
        """
        Update information on the account. Note that many fields may be unalterable
        by the API key holder.

        :calls: "PUT /accounts/{id}" \
		<https://fabman.io/api/v1/documentation#/account/putAccountsId>

        :returns: None
        :rtype: None
        """
        uri = f"/accounts/{self.id}"

        kwargs.update({"lockVersion": self.lockVersion})

        response = self._requester.request("PUT", uri, _kwargs=kwargs)

        data = response.json()

        for attr, val in data.items():
            setattr(self, attr, val)

    def get_payment_info(self, **kwargs) -> PaymentInfo:
        """
        Get information about the payment plan of the account.

        :calls: "GET /accounts/{id}/payment-info" \
        <https://fabman.io/api/v1/documentation#/accounts/getAccountsIdPaymentinfo>
        
        :returns: Information about the payment plan of the account.
        :rtype: fabman.account.PaymentInfo
        """
        uri = f"/accounts/{self.id}/payment-info"

        response = self._requester.request("GET", uri, _kwargs=kwargs)
        data = response.json()
        data.update({"account_id": self.id})

        return PaymentInfo(self._requester, data)



"""Handles pagination of the api"""
from typing import List, Optional, Type

from requests.structures import CaseInsensitiveDict

from fabman.fabman_object import FabmanObject
from fabman.requester import Requester


# TODO: kwargs are not coming through correctly, so commands for pagination are not working
class PaginatedList(object):  # pylint: disable=too-many-instance-attributes
    """
    Abstracts pagination and rate limiting of the Fabman API.
    Documentation: https://github.com/FabmanHQ/fabman-api#pagination
    """

    def __getitem__(self, index):
        assert isinstance(
            index, int
        ), "Index must be an integer, slicing and keys are \
            not supported"
        if index < 0:
            raise IndexError("Cannot use negative indexing on PaginatedList")
        self._get_up_to_index(index)
        return self._elements[index]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        content_class: Type[FabmanObject],
        requester: Requester,
        request_method: str,
        first_url: str,
        extra_attribs: Optional[dict] = None,
        _root: Optional[str] = None,
        url_override: Optional[str] = None,
        **kwargs,
    ) -> None:
        """Abstracts pagination of the Fabman API. Provides a simple interface to work with
        objects and requests new objects from the API as needed.

        :param content_class: Class of the stored objects. Must inherit from FabmanObject
        :type content_class: Type[fabman.fabman_object.FabmanObject]
        :param requester: Requester for refreshing the API
        :type requester: fabman.requester.Requester
        :param request_method: One of "GET", "POST", "PUT", "DELETE"
        :type request_method: str
        :param first_url: endpoint to be called for the first page
        :type first_url: str
        :param extra_attribs: Extra attributes to be added to :code:`content_class`, defaults to None
        :type extra_attribs: Optional[dict], optional
        :param _root: Root endpoint without parameters, defaults to None
        :type _root: str, optional
        :param url_override: Override the base_url, defaults to None
        :type url_override: str, optional
        """

        self._elements = []

        self._content_class = content_class
        self._requester = requester
        self._first_url = first_url
        self._first_params = kwargs or {}
        self._next_url = first_url
        self._next_params = self._first_params
        self._extra_attribs = extra_attribs or {}
        self._root = _root
        self._request_method = request_method
        self._url_override = url_override

    def __iter__(self):
        for element in self._elements:
            yield element
        while self._has_next():
            new_elements = self._grow()
            for element in new_elements:
                yield element

    def __repr__(self):
        return f"<PaginatedList of type {self._content_class.__name__}>"

    @staticmethod
    def __format_link(headers: CaseInsensitiveDict):
        if "link" in headers.keys():
            link = headers["link"]
            return link.split(";")[0].strip("<>").split("/api/v1")[1]

        return None

    def _get_next_page(self):
        response = self._requester.request(
            self._request_method,
            self._next_url,
            _url=self._url_override,
            _kwargs=self._next_params,
        )

        data = response.json()
        headers = response.headers
        self._next_url = self.__format_link(headers)

        content = []

        for element in data:
            if element is not None:
                element.update(self._extra_attribs)
                content.append(self._content_class(self._requester, element))

        return content

    def _get_up_to_index(self, index):
        while len(self._elements) <= index and self._has_next():
            self._grow()

    def _grow(self):
        new_elements = self._get_next_page()
        self._elements.extend(new_elements)
        return new_elements

    def _has_next(self):
        return self._next_url is not None



import re
from distutils.core import setup
from os import path

# get version number from fabman/__init__.py
with open("fabman/__init__.py") as f:
    version = re.search(r"__version__ = (.*)", f.read()).group(1)

setup(
    name="Fabman API",
    version=version,
    description="Library for interfacing with the Fabman API",
    author="Davin Lawrence",
    author_email="fabman-api@tinycact.us",
    packages=["fabman"],
    install_requires=["requests"],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Education",
    ],
)



"""Interface for handling embedded lists of FabmanObjects"""

from typing import List, Optional, Type

from fabman.fabman_object import FabmanObject
from fabman.paginated_list import PaginatedList
from fabman.requester import Requester


class EmbeddedList(object):
    """Mimics the functionality of :code:`fabman.paginated_list.PaginatedList` for embedded objects."""

    def __getitem__(self, index):
        assert isinstance(
            index, int
        ), "Index must be an integer. slicing and keys are not supported"
        if index < 0:
            raise IndexError("Cannot use negative indexing on EmbeddedList")
        return self._elements[index]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        content_class: Type[FabmanObject],
        initial_data: List[dict],
        requester: Requester,
        request_method: Optional[str] = None,
        refresh_endpoint: Optional[str] = None,
        extra_attribs: Optional[dict] = None,
        **kwargs,
    ) -> None:
        """Initializes the EmbeddedList object with sub element and acts as a simple list.
        Has a refresh method to retrieve a PaginatedList from the API.

        :param content_class: Class of the stored objects.
        :type content_class: Type[FabmanObject]
        :param requester: Requester for refreshing the API
        :type requester: fabman.requester.Requester
        :param request_method: One of "GET", "POST", "PUT", "DELETE"
        :type request_method: str
        :param refresh_endpoint: Endpoint for retrieving a Paginated List
        :type refresh_endpoint: str
        :param initial_data: list of Initial data in JSON format
        :type initial_data: List[dict]
        :param extra_attribs: Extra attributes to add to each object, defaults to None
        :type extra_attribs: Optional[dict], optional
        """
        self._elements = []

        for element in initial_data:
            if extra_attribs:
                element.update(extra_attribs)
            obj = content_class(requester, element)
            self._elements.append(obj)

        self._content_class = content_class
        self._requester = requester
        self._first_params = kwargs or {}
        self._refresh_endpoint = refresh_endpoint
        self._extra_attribs = extra_attribs or {}
        self._request_method = request_method

    def __iter__(self):
        for element in self._elements:
            yield element

    def __repr__(self):
        return f"<EmbeddedList of type {self._content_class.__name__} with {len(self._elements)} elements>"

    def get_live_data(self) -> PaginatedList:
        """Convert the EmbeddedList into a PaginatedList by passing arguments to the
        PaginatedList. The Paginated list will hold no data but will retrieve live
        data as needed.

        :return: PaginatedList to work on current objects
        :rtype: PaginatedList
        """
        if self._refresh_endpoint is None or self._refresh_endpoint == "":
            raise ValueError("Cannot refresh an EmbeddedList without a refresh URL")
        if self._request_method is None or self._request_method == "":
            raise ValueError("Cannot refresh an EmbeddedList without a request method")

        return PaginatedList(
            self._content_class,
            self._requester,
            self._request_method,
            self._refresh_endpoint,
            extra_attribs=self._extra_attribs,
            **self._first_params,
        )
